# configs/__init__.py

import os

# Define the path to the linkage.json file
LINKAGE_CONFIG_PATH = os.path.join(os.path.dirname(__file__), "linkage.json")
CLF_CONFIG_PATH = os.path.join(os.path.dirname(__file__), "classification.json")
